import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.FontMetrics;
import java.util.Arrays;

/**
 * @author David Foster <davidfstr@gmail.com>
 */
public class GrayscaleLevelHistogram implements PixelBufferToBufferedImageFilter, ParameterizedFilter {
	private final static int HISTOGRAM_BIN_WIDTH = 1;
	
	private final static int HISTOGRAM_WIDTH = HISTOGRAM_BIN_WIDTH * 256;
	private final static int HISTOGRAM_HEIGHT = 150;
	
	private final static Color BKGND_COLOR = Color.WHITE;
	private final static Color FRGND_COLOR = new Color(255*1/2, 255*1/2, 255);
	private final static Color BORDER_COLOR = Color.BLACK;
	private final static Color OPTIMAL_THRESHOLD_COLOR = Color.RED;
	private final static Color STATISTICS_COLOR = Color.BLUE;
	
	private final static Color[] BIN_COLORS;
	static {
		BIN_COLORS = new Color[256];
		for (int i=0; i<256; i++) {
			BIN_COLORS[i] = new Color(i, i, i);
		}
	}
	
	private final FloatFilterParam binHeightScaleFactorParam =
		new FloatFilterParam("Bin Height Scale Factor", 0.00f, Float.MAX_VALUE, 1.00f);
	
	private final FilterParam[] params = new FilterParam[] {
		binHeightScaleFactorParam
	};
	
	// ### FILTER ###
	
	private int[] levels = new int[256];
	private int[] inGrayPixels = null;
	
	public void filter(int width, int height, int[] inPixels,
	                   BufferedImage inImg, BufferedImage outImg)
	{
		float binHeightScaleFactor = binHeightScaleFactorParam.getValue();
		
		int numPixels = inPixels.length;
		
		/*
		 * Obtain grayscale image from input image
		 */
		{
			if ((inGrayPixels == null) || (inGrayPixels.length != numPixels)) {
				inGrayPixels = new int[numPixels];
			}
			
			for (int i=0; i<numPixels; i++) {
				inGrayPixels[i] = pixelToLevel(inPixels[i]);
			}
		}
		
		/* 
		 * Obtain histogram of the frequencies of the 
		 * different levels of gray in the input image
		 */
		int maxLevel = 0;
		{
			for (int bin=0; bin<256; bin++) {
				levels[bin] = 0;
			}
			
			for (int i=0; i<numPixels; i++) {
				levels[inGrayPixels[i]]++;
			}
			
			for (int bin=0; bin<256; bin++) {
				int curLevel = levels[bin];
				if (curLevel > maxLevel) {
					maxLevel = curLevel;
				}
			}
		}
		
		/*
		 * Calculate optimal threshold value using
		 * the Otsu auto-thresholding algorithm
		 */
		int optimalThreshold = 
			OtsuThresholdingAlgorithm.
				calculateOptimalThresholdWithHistogram(
					levels /* histogram */,
					numPixels);
		
		/*
		 * Calculate statistics related to the determined threshold value
		 * 
		 * Classes:
		 * - C1: [0,optimalThreshold]
		 * - C2: [optimalThreshold+1,255]
		 * - C2*: [optimalThreshold,255]
		 */
		int percentAreaC1, percentAreaC2;
		int meanC1, meanC2;
		int stddevC1, stddevC2;
		int percentileOfThresholdFreqInC2Freqs;		// *might* be useful for determining whether a barrel is present in the input image
		{
			/* Calculate C1 stats */
			{
				//int N = (optimalThreshold - 0) + 1;
				
				int areaC1 = 0;
				for (int i=0; i<=optimalThreshold; i++) {
					areaC1 += levels[i];
				}
				percentAreaC1 = areaC1*100/numPixels;
				
				meanC1 = 0;
				for (int i=0; i<=optimalThreshold; i++) {
					meanC1 += levels[i]*i;
				}
				meanC1 /= areaC1;
				
				int sumSquareDevs = 0;
				for (int i=0; i<=optimalThreshold; i++) {
					sumSquareDevs += levels[i]*square(i - meanC1);
				}
				stddevC1 = (int) Math.sqrt(((double) sumSquareDevs)/(areaC1-1));
			}
			
			/* Calculate C2 stats */
			{
				//int N = (255 - (optimalThreshold+1)) + 1;
				
				int areaC2 = 0;
				for (int i=optimalThreshold+1; i<256; i++) {
					areaC2 += levels[i];
				}
				percentAreaC2 = areaC2*100/numPixels;
				
				meanC2 = 0;
				for (int i=optimalThreshold+1; i<256; i++) {
					meanC2 += levels[i]*i;
				}
				meanC2 /= areaC2;
				
				int sumSquareDevs = 0;
				for (int i=optimalThreshold+1; i<256; i++) {
					sumSquareDevs += levels[i]*square(i - meanC2);
				}
				stddevC2 = (int) Math.sqrt(((double) sumSquareDevs)/(areaC2-1));
			}
			
			/* 
			 * Calculate percentile of threshold-frequency relative to the
			 * frequencies of levels in class C2
			 */
			{
				int[] freqsInC2 = new int[256-optimalThreshold];
				for (int i=optimalThreshold, j=0; i<256; i++, j++) {
					freqsInC2[j] = levels[i];
				}
				
				Arrays.sort(freqsInC2);
				
				int pos = Arrays.binarySearch(freqsInC2, levels[optimalThreshold]);
				if (pos < 0)
					throw new IllegalStateException();
				percentileOfThresholdFreqInC2Freqs = pos*100/freqsInC2.length;
			}
		}
		
		/*
		 * Begin drawing
		 */
		Graphics2D g = (Graphics2D) outImg.getGraphics();
		{
			/* Fill output image with background color behind histogram */
			g.setColor(BKGND_COLOR);
			g.fillRect(0, 0, width, height);
		}
		
		/*
		 * Print statistics related to the determined threshold value
		 */
		{
			String[] allLines = new String[] {
				"Threshold (t*): "+optimalThreshold+"/255",
				"% in C1: "+percentAreaC1+"%",
				"% in C2: "+percentAreaC2+"%",
				"Mean C1: "+meanC1,
				"Mean C2: "+meanC2,
				"Stddev C1: "+stddevC1,
				"Stddev C2: "+stddevC2,
				"%ile of t* freq within C2* freqs: "+percentileOfThresholdFreqInC2Freqs
			};
			
			FontMetrics fontMetrics = g.getFontMetrics();
			g.setColor(STATISTICS_COLOR);
			
			int distFromTopToBaseline = fontMetrics.getAscent();
			int distBetweenLines = fontMetrics.getHeight();
			for (int i=0, n=allLines.length, y=distFromTopToBaseline;
			     i<n;
			     i++, y+=distBetweenLines)
			{
				g.drawString(allLines[i], 0, y);
			}
		}
		
		/* Draw the histogram */
		{
			/* 
			 * Translate coordinate space to the upper-left corner of
			 * the histogram (which is centered in the middle of the output image)
			 */
			g.translate(
				(width - HISTOGRAM_WIDTH)/2,
				(height - HISTOGRAM_HEIGHT)/2);
			
			/* Draw border around histogram */
			g.setColor(BORDER_COLOR);
			g.drawRect(-1, -1, HISTOGRAM_WIDTH+1, HISTOGRAM_HEIGHT+1);
			
			/* Draw foreground of histogram */
			g.setColor(FRGND_COLOR);
			g.fillRect(0, 0, HISTOGRAM_WIDTH, HISTOGRAM_HEIGHT);
			
			/* Clip subsequent drawing inside the histogram region */
			g.setClip(0, 0, HISTOGRAM_WIDTH, HISTOGRAM_HEIGHT);
			
			/* Draw the levels inside the histogram */
			int binX=0;
			for (int bin=0; bin<256; bin++, binX+=HISTOGRAM_BIN_WIDTH) {
				int curLevel = levels[bin];
				int curBinHeight = (int) (binHeightScaleFactor*(curLevel*HISTOGRAM_HEIGHT)/maxLevel);
				Color curBinColor = BIN_COLORS[bin];
				
				// Draw the current bin
				g.setColor(curBinColor);
				g.fillRect(
					binX, HISTOGRAM_HEIGHT-curBinHeight,
					HISTOGRAM_BIN_WIDTH, curBinHeight);
			}
			
			/* Draw the calculated optimal threshold value */
			g.setColor(OPTIMAL_THRESHOLD_COLOR);
			g.drawLine(optimalThreshold, 0, optimalThreshold, HISTOGRAM_HEIGHT-1);
		}
	}
	
	private static int pixelToLevel(int pixel) {
		return (Pixels.getRed(pixel) + Pixels.getGreen(pixel) + Pixels.getBlue(pixel))/3;
	}
	
	private static int square(int n) {
		return n*n;
	}
	
	public String toString() {
		return "Grayscale Level Histogram";
	}
	
	// ### PARAMETERIZED FILTER ###
	
	public FilterParam[] getParams() {
		return params;
	}
}